from typing import List, Union

from pydantic import BaseModel, Field, create_model, ConfigDict

from fridonai_core.plugins import BasePlugin


def create_plugin_wrapper_tool(plugin: BasePlugin, class_name: str) -> type[BaseModel]:
    return create_model(
        f"To{class_name}",
        __config__=type(
            "Config",
            (ConfigDict,),
            {
                "json_schema_extra": {
                    "Example#" + str(i + 1): {"request": example}
                    for i, example in enumerate(plugin.request_examples)
                }
            },
        ),
        __doc__="Transfer control to the assistant: " + plugin.full_description(),
        request=(str, Field(description="Full requests from the user.")),
    )


class CompleteTool(BaseModel):
    """A tool to mark the current agent's work as completed with appropriate answer and return control to the main assistant,
    who can re-route the dialog based on the user's needs. Tool doesn't make up any information and is 100% precise with generating answer.
    If the agent's response was json object answer must be a dict otherwise it must be a string."""

    plugin_status: bool = Field(
        description="Plugin made it's job successfully, or not and something wrong happened"
    )
    answer: str = Field(
        description="Plugin's response, if json string is the response copy it as it is don't cut or add something."
    )
    is_json: bool = Field(description="If the answer is json string or not.")

    class Config:
        json_schema_extra = {
            "example": {
                "plugin_status": True,
                "answer": "Your sol balance is 20.",
            },
            "example 2": {
                "plugin_status": True,
                "answer": "You've successfully borrowed 10 sol on kamino",
            },
            "example 3": {"plugin_status": False, "answer": "Something went wrong!"},
            "example 4": {
                "plugin_status": False,
                "answer": {"data": {"sol": 12, "usdc": 10}},
            },
        }


class FinalResponse(BaseModel):
    """FinalResponse generator tool, in no other tools is going to be called this tool is called to generate the final answers.
    A tool to generate final response to the user's request. It must be the last tool call in the conversation with the final answer."""

    structured_answers: Union[List[Union[str, dict]], None] = Field(
        description="All json string answers from `CompletedTool`s should be copied here if they are json."
    )
    text_answer: Union[str, None] = Field(
        description="Union of tool text answers from `CompleteTool`s tools which wasn't they aren't json."
    )

    class Config:
        json_schema_extra = {
            "example": {
                "structured_answers": [
                    '"data": {"sol": 12, "usdc": 10}}',
                    '"data": {"sol": 12, "usdc": 10}',
                ],
                "text_answer": "Your transaction successfully completed and you also successfully staked 10 sol on kamino.",
            },
            "example 2": {"text_answer": "Solana looks very decent to be bought."},
            "example 3": {
                "structured_answer": [
                    '"data": {"btc": {"score": 1',
                    '"eth": {"score": 0}}',
                ]
            },
        }



import json
from typing import Any
import uuid

from langchain.callbacks.manager import AsyncCallbackManager, CallbackManagerForToolRun
from langchain_core.runnables import RunnableConfig, ensure_config
from langchain_core.tools import BaseTool as LangchainBaseTool

from fridonai_core.plugins.schemas import BaseToolInput
from fridonai_core.plugins.utilities import BaseUtility


class BaseTool(LangchainBaseTool):
    name: str
    description: str
    args_schema: type[BaseToolInput] | None = None
    utility: BaseUtility
    examples: list[dict] = []
    helper: bool = False
    dump_json_data: bool = False

    def _run(
        self,
        *args,
        config: RunnableConfig | None = None,
        run_manager: CallbackManagerForToolRun | None = None,
        **kwargs,
    ) -> str | Any:
        config_ = config.get("configurable", {})
        result = self.utility.run(
            *args,
            **{**kwargs, **config_},
        )
        if self.dump_json_data and isinstance(result, dict):
            result = self._dump_to_tmp(result)
        elif isinstance(result, dict):
            result = json.dumps(result)
        return result

    async def _arun(
        self,
        *args,
        config: RunnableConfig | None = None,
        run_manager: AsyncCallbackManager | None = None,
        **kwargs,
    ) -> str | Any:
        config = ensure_config()
        config_ = config.get("configurable", {})
        result = await self.utility.arun(
            *args,
            **{**kwargs, **config_},
        )
        if self.dump_json_data and isinstance(result, dict):
            result = self._dump_to_tmp(result)
        return result
    

    def _dump_to_tmp(self, result: str | Any):
        temp_file_id = str(uuid.uuid4())[:8]
        try:
            with open(f"tmp/{temp_file_id}.json", 'w') as f:
                json.dump(result, f, ensure_ascii=False, indent=2)
        except Exception as e:
            return result

        return json.dumps({
            "type": "fileObject",
            "name": self.name,
            "tmp_file_id": temp_file_id,
            "path": f"tmp/{temp_file_id}.json",
        })
        

    @property
    def parameters_as_str(self):
        return self.args_schema.class_parameters_as_string()

    def to_json(self):
        return {
            "name": self.name,
            "description": self.description,
            "examples": self.examples,
            "parameters": self.parameters_as_str,
        }



from contextvars import ContextVar

from pydantic import BaseModel

from fridonai_core.crons.base import BaseCron


class CronRegistry(BaseModel):
    crons: list[BaseCron] = []

    def register(self, cron: type[BaseCron]) -> type[BaseCron]:
        # print(cron)
        self.crons.append(cron())
        return cron

    async def start_crons(self) -> None:
        # print(self.crons)
        for cron in self.crons:
            await cron.start()


var_cron_registry = ContextVar("cron_registry", default=CronRegistry())


def ensure_cron_registry() -> CronRegistry:
    return var_cron_registry.get()



from datetime import datetime

from aiocron import crontab
from pydantic import BaseModel


class BaseCron(BaseModel):
    name: str
    schedule: str

    async def arun(self) -> None:
        print(f"Running cron job {self.name} - {datetime.now()}")
        await self._process()

    async def start(self) -> None:
        await self.arun()
        crontab(self.schedule, self.arun)

    async def _process(self) -> None: ...



from typing import Any, Optional

import aiohttp
from pydantic import Field

from fridonai_core.plugins.utilities.adapter_base import BaseAdapter
from fridonai_core.plugins.utilities.base import BaseUtility


class BlockchainUtility(BaseUtility):
    request_url: str = Field(
        default="http://localhost:3000/data/executor",
        exclude=True,
    )

    communicator: Optional[BaseAdapter] = Field(default=None, exclude=True)

    async def _generate_tx(self, request: dict[str, Any]) -> str | dict | Any:
        async with aiohttp.ClientSession() as session:
            async with session.post(self.request_url, json=request) as resp:
                resp = await resp.json()
                if "statusCode" in resp:
                    if 500 > resp["statusCode"] >= 400:
                        return resp.get("message", "Something went wrong!")
                    if resp["statusCode"] >= 500:
                        return "Something went wrong! Please try again later."

                return resp["data"]["serializedTx"]

    async def _send_and_wait(
        self,
        tx: dict | None,
        wallet_id: str,
        chat_id: str,
    ) -> str:
        response = await self.communicator.send(
            "response_received",
            {
                "chat_id": chat_id,
                "wallet_id": wallet_id, 
                "tx": tx,
            }
        )
        print("Got Response", response)
        return response

    async def arun(self, *args, chat_id: str, **kwargs) -> str:
        request = await self._arun(*args, **kwargs)

        tx = await self._generate_tx(request)
        result = await self._send_and_wait(tx, kwargs.get("wallet_id", ""), chat_id)

        return result

    class Config:
        arbitrary_types_allowed = True



import os 
from typing import Literal, Union
from langchain_core.language_models import BaseChatModel
from langchain_core.runnables import RunnableLambda
from langgraph.checkpoint.base import BaseCheckpointSaver
from langgraph.graph import END, StateGraph
from langgraph.prebuilt import ToolNode
from langchain_core.messages import HumanMessage

from fridonai_core.graph.agents import create_agent
from fridonai_core.graph.chains import create_agent_chain
from fridonai_core.graph.prompts import create_agent_prompt, create_supervised_prompt
from fridonai_core.graph.routers import route_plugin_agent, route_supervisor_agent
from fridonai_core.graph.states import State
from fridonai_core.graph.tools import CompleteTool, create_plugin_wrapper_tool
from fridonai_core.graph.utils import (
    generate_structured_response,
    handle_tool_error,
    leave_tool,
    prepare_plugin_agent,
)
from fridonai_core.plugins import BasePlugin


def create_graph(
    llm: BaseChatModel, plugins: list[BasePlugin], memory: BaseCheckpointSaver
):
    plugins_to_wrapped_plugins = {
        p.name: create_plugin_wrapper_tool(p, type(p).__name__) for p in plugins
    }
    wrapped_plugins_to_plugins = {
        wp.__name__: pname for pname, wp in plugins_to_wrapped_plugins.items()
    }

    workflow = StateGraph(State)

    supervisor_agent = create_agent(
        create_agent_chain(
            llm,
            prompt=create_supervised_prompt(),
            tools=[*list(plugins_to_wrapped_plugins.values())],
            always_tool_call=False,
        )
    )

    workflow.add_node("supervisor", supervisor_agent)

    for plugin in plugins:
        agent_chain = create_agent_chain(
            llm,
            create_agent_prompt(
                plugin.name,
                plugin.full_description(tool_descriptions=False),
                plugin.output_format,
            ),
            tools=plugin.tools + [CompleteTool],
            always_tool_call=True,
        )

        tool_node = ToolNode(plugin.tools + [CompleteTool]).with_fallbacks(
            [RunnableLambda(handle_tool_error)], exception_key="error"
        )

        agent = create_agent(agent_chain, plugin.slug)

        workflow.add_node("Enter" + plugin.name, prepare_plugin_agent)
        workflow.add_node(plugin.name, agent)
        workflow.add_edge("Enter" + plugin.name, plugin.name)
        workflow.add_node(plugin.name + "_tools", tool_node)
        workflow.add_edge(plugin.name + "_tools", plugin.name)
        workflow.add_conditional_edges(
            plugin.name,
            route_plugin_agent,
            {
                "tool_node": plugin.name + "_tools",
                "leave_tool": "leave_tool",
                "supervisor": "supervisor",
            },
        )

    def _route_supervisor_agent_wrapper(state):
        return route_supervisor_agent(state, wrapped_plugins_to_plugins)

    workflow.add_conditional_edges(
        "supervisor",
        _route_supervisor_agent_wrapper,
        {
            **{"Enter" + plugin.name: "Enter" + plugin.name for plugin in plugins},
            "respond": "respond",
        },
    )
    workflow.add_node("respond", generate_structured_response)
    workflow.add_node("leave_tool", leave_tool)
    workflow.add_edge("leave_tool", "supervisor")
    workflow.add_edge("respond", END)

    workflow.set_entry_point("supervisor")

    return workflow.compile(checkpointer=memory)


async def generate_response(
    message: str, 
    plugins: list[BasePlugin],
    llm: BaseChatModel,
    config: dict,
    memory: Literal['postgres'] = 'postgres',
    return_used_agents: bool = True,
):
        
    if memory not in ['postgres', 'sqlite']:
        raise ValueError('Only postgres and sqlite are supported for now')

    if memory == 'sqlite':
        from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver as Saver
        conn_string = ":memory:"
    else: 
        from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver as Saver
        conn_string = os.environ.get("POSTGRES_DB_URL")

    async with Saver.from_conn_string(conn_string) as memory:
        if memory == 'postgres':
            await memory.setup()

        graph = create_graph(llm, plugins, memory)
        graph_config = {"configurable": config}
        graph_state = await invoke_graph(message, graph, graph_config)
        
        final_response = graph_state.values.get("final_response")
        if return_used_agents:
            used_agents = list(set(graph_state.values.get("used_agents", [])))
            return final_response, used_agents
        return final_response



async def invoke_graph(message, graph, graph_config):
    response = ""
    async for s in graph.astream(
        {
            "messages": [HumanMessage(content=message)],
        },
        graph_config,
        stream_mode="values",
    ):
        if "__end__" not in s:
            response = s["messages"][-1]
            response.pretty_print()
            graph_state = await graph.aget_state(graph_config)

        return graph_state


from typing import Union
from langchain_core import BaseModel, Field


class ResponseSchema(BaseModel):
    plugin_status: bool = Field(
        description="Plugin made it's job successfully, or not and something wrong happened"
    )
    answer_text: str = Field(description="Text response from the plugin")
    structured_response: Union[dict, BaseModel] = Field(
        description="Structured response from the plugin"
    )



from langgraph.graph import END
from langgraph.prebuilt import tools_condition

from fridonai_core.graph.states import State
from fridonai_core.graph.tools import CompleteTool


def route_plugin_agent(
    state,
) -> str:
    route = tools_condition(state)
    if route == END:
        return "supervisor"
    tool_calls = state["messages"][-1].tool_calls
    did_cancel = any(tc["name"] == CompleteTool.__name__ for tc in tool_calls)
    if did_cancel:
        return "leave_tool"
    return "tool_node"


def route_supervisor_agent(state: State, wrapped_plugins_to_plugins) -> str:
    tool_calls = state["messages"][-1].tool_calls
    if not tool_calls:
        return "respond"
    return "Enter" + wrapped_plugins_to_plugins[tool_calls[0]["name"]]



from fridonai_core.plugins.utilities.base import BaseUtility
from fridonai_core.plugins.utilities.blockchain import BlockchainUtility
from fridonai_core.plugins.utilities.remote import RemoteUtility
from fridonai_core.plugins.utilities.mock import *
from fridonai_core.plugins.utilities.llm import LLMUtility

__all__ = ["BaseUtility", "BlockchainUtility", "RemoteUtility", "MockUtility", "LLMUtility"]



from pydantic import BaseModel


class BaseToolInput(BaseModel):
    @classmethod
    def class_parameters_as_string(cls) -> str:
        explanation = []
        for field_name, field_info in cls.model_fields.items():
            description = field_info.description or ""
            explanation.append(f"{field_name} - {description}")
        return "\n".join(explanation)


from langchain_core.runnables import Runnable, RunnableConfig


class Agent:
    def __init__(self, runnable: Runnable, name: str = "Agent"):
        self.name = name
        self.runnable = runnable

    def __call__(self, state, config: RunnableConfig):
        while True:
            result = self.runnable.invoke(state)

            if not result.tool_calls and (
                not result.content
                or isinstance(result.content, list)
                and not result.content[0].get("text")
            ):
                messages = state["messages"] + [("user", "Respond with a real output.")]
                state = {**state, "messages": messages}
            else:
                break
        return {
            "messages": result,
            "used_agents": [self.name] if self.name != "Agent" else [],
        }


class SupervisorAgent(Agent):
    def __call__(self, state, config: RunnableConfig):
        result = super().__call__(state, config)["messages"]
        grouped_tool_calls = {}

        for tc in result.tool_calls:
            tool_name = tc["name"]
            if tool_name not in grouped_tool_calls:
                grouped_tool_calls[tool_name] = {**tc, "args": {}}
            for key in tc["args"]:
                if key not in grouped_tool_calls[tool_name]["args"]:
                    grouped_tool_calls[tool_name]["args"][key] = tc["args"][key]
                else:
                    grouped_tool_calls[tool_name]["args"][key] += tc["args"][key]

        tcs = list(grouped_tool_calls.values())

        result.tool_calls = [] if len(tcs) == 0 else [tcs[0]]

        return {"messages": result, "used_agents": []}


def create_agent(runnable: Runnable, name: str = "Agent") -> Agent:
    if name == "Agent":
        return SupervisorAgent(runnable, "Agent")
    return Agent(runnable, name)



from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable
from pydantic import BaseModel

from fridonai_core.plugins.tools.base import BaseTool


def create_agent_chain(
    llm: BaseChatModel,
    prompt: ChatPromptTemplate,
    tools: list[BaseTool | type[BaseModel]],
    always_tool_call=False,
) -> Runnable:
    if always_tool_call:
        return prompt | llm.bind_tools(
            tools, tool_choice="any", parallel_tool_calls=False
        )
    return prompt | llm.bind_tools(tools, parallel_tool_calls=False)



import importlib
import pkgutil
from contextvars import ContextVar
from types import ModuleType

from pydantic import BaseModel

from fridonai_core.plugins import BasePlugin


class PluginRegistry(BaseModel):
    plugins: dict[str, type[BasePlugin]] = {}

    @property
    def plugins_list(self) -> list[type[BasePlugin]]:
        return list(self.plugins.values())

    def register(self, name: str) -> callable:
        def wrapper(plugin: type[BasePlugin]):
            self.plugins[name] = plugin
            return plugin

        return wrapper

var_plugin_registry = ContextVar("plugin_registry", default=PluginRegistry())


def ensure_plugin_registry() -> PluginRegistry:
    return var_plugin_registry.get()


def resolve_module_plugins(module: ModuleType):
    for _, module_name, _ in pkgutil.walk_packages(module.__path__):
        importlib.import_module(f"{module.__name__}.{module_name}")



from operator import add
from typing import Annotated, List, TypedDict, Union

from langchain_core.messages import AnyMessage
from langgraph.graph import add_messages

from fridonai_core.graph.tools import FinalResponse


class State(TypedDict):
    messages: Annotated[List[AnyMessage], add_messages]
    used_agents: Annotated[list[str], add]
    final_response: Union[FinalResponse, None]



from typing import Any

from pydantic import BaseModel


class BaseUtility(BaseModel):
    def run(self, *args, **kwargs) -> dict | str | Any: ...

    def _run(self, *args, **kwargs) -> str: ...

    async def arun(self, *args, **kwargs) -> dict | str | Any: ...

    async def _arun(self, *args, **kwargs) -> dict | str: ...
